###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os
import sys
import unittest
from unittest import mock

from autobahn.nvx import _compile_args
from autobahn.nvx._compile_args import (
    get_compile_args,
    _get_safe_march_flag,
    _get_target_machine,
)


class TestSafeMarchFlag(unittest.TestCase):
    """Architecture -> safe -march flag mapping (#1834, #1717)."""

    def test_x86_64_variants(self):
        for machine in ("x86_64", "amd64", "x64"):
            self.assertEqual(_get_safe_march_flag(machine), None)

    def test_arm64_variants(self):
        for machine in ("aarch64", "arm64"):
            self.assertEqual(_get_safe_march_flag(machine), "-march=armv8-a")

    def test_unknown_arch_returns_none(self):
        # Unknown architectures must not get a -march flag (let the toolchain
        # defaults / CFLAGS decide), so cross-compilation is never broken.
        self.assertIsNone(_get_safe_march_flag("riscv64"))


class TestTargetMachine(unittest.TestCase):
    def test_returns_nonempty_lowercase(self):
        machine = _get_target_machine()
        self.assertIsInstance(machine, str)
        self.assertTrue(machine)
        self.assertEqual(machine, machine.lower())

    def test_uses_sysconfig_target_arch(self):
        # When cross-compiling, sysconfig.get_platform() reflects the *target*
        # (e.g. aarch64) even on an x86-64 build host - this is the whole point
        # of preferring it over platform.machine() (#1834 / PR #1835).
        with mock.patch(
            "autobahn.nvx._compile_args.sysconfig.get_platform",
            return_value="linux-aarch64",
        ):
            self.assertEqual(_get_target_machine(), "aarch64")


class TestGetCompileArgs(unittest.TestCase):
    """Default-safe / opt-in-native behaviour (#1834)."""

    def test_default_is_never_native(self):
        # The core #1834 guard: with no override, -march=native must NEVER be
        # emitted (it breaks cross-compilation and can SIGILL distributed wheels).
        with mock.patch.dict("os.environ", {}, clear=False):
            os.environ.pop("AUTOBAHN_ARCH_TARGET", None)
            self.assertNotIn("-march=native", get_compile_args())

    def test_safe_matches_default(self):
        with mock.patch.dict("os.environ", {}, clear=False):
            os.environ.pop("AUTOBAHN_ARCH_TARGET", None)
            default_args = get_compile_args()
        with mock.patch.dict("os.environ", {"AUTOBAHN_ARCH_TARGET": "safe"}):
            self.assertEqual(get_compile_args(), default_args)

    @unittest.skipIf(sys.platform == "win32", "MSVC path uses /arch, not -march")
    def test_native_is_opt_in(self):
        with mock.patch.dict("os.environ", {"AUTOBAHN_ARCH_TARGET": "native"}):
            self.assertIn("-march=native", get_compile_args())

    @unittest.skipIf(sys.platform == "win32", "MSVC path uses /arch, not -march")
    def test_unknown_target_emits_no_march(self):
        # An unrecognized cross-compile target must yield no -march flag at all.
        with mock.patch.dict("os.environ", {}, clear=False):
            os.environ.pop("AUTOBAHN_ARCH_TARGET", None)
            with mock.patch.object(
                _compile_args, "_get_target_machine", return_value="riscv64"
            ):
                args = get_compile_args()
        self.assertFalse(any(a.startswith("-march") for a in args))

    @unittest.skipIf(sys.platform != "win32", "MSVC-specific flags")
    def test_windows_uses_msvc_flags(self):
        self.assertEqual(get_compile_args(), ["/O2", "/W3"])


if __name__ == "__main__":
    unittest.main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class ClientRoles(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ClientRoles()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsClientRoles(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # ClientRoles
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ClientRoles
    def Publisher(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from autobahn.wamp.gen.wamp.proto.PublisherFeatures import PublisherFeatures

            obj = PublisherFeatures()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ClientRoles
    def Subscriber(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from autobahn.wamp.gen.wamp.proto.SubscriberFeatures import (
                SubscriberFeatures,
            )

            obj = SubscriberFeatures()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ClientRoles
    def Caller(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from autobahn.wamp.gen.wamp.proto.CallerFeatures import CallerFeatures

            obj = CallerFeatures()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ClientRoles
    def Callee(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from autobahn.wamp.gen.wamp.proto.CalleeFeatures import CalleeFeatures

            obj = CalleeFeatures()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None


def ClientRolesStart(builder):
    builder.StartObject(4)


def Start(builder):
    ClientRolesStart(builder)


def ClientRolesAddPublisher(builder, publisher):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(publisher), 0
    )


def AddPublisher(builder, publisher):
    ClientRolesAddPublisher(builder, publisher)


def ClientRolesAddSubscriber(builder, subscriber):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(subscriber), 0
    )


def AddSubscriber(builder, subscriber):
    ClientRolesAddSubscriber(builder, subscriber)


def ClientRolesAddCaller(builder, caller):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(caller), 0
    )


def AddCaller(builder, caller):
    ClientRolesAddCaller(builder, caller)


def ClientRolesAddCallee(builder, callee):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(callee), 0
    )


def AddCallee(builder, callee):
    ClientRolesAddCallee(builder, callee)


def ClientRolesEnd(builder):
    return builder.EndObject()


def End(builder):
    return ClientRolesEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class AuthScramWelcome(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = AuthScramWelcome()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAuthScramWelcome(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # AuthScramWelcome
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # AuthScramWelcome
    def Verifier(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def AuthScramWelcomeStart(builder):
    builder.StartObject(1)


def Start(builder):
    AuthScramWelcomeStart(builder)


def AuthScramWelcomeAddVerifier(builder, verifier):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(verifier), 0
    )


def AddVerifier(builder, verifier):
    AuthScramWelcomeAddVerifier(builder, verifier)


def AuthScramWelcomeEnd(builder):
    return builder.EndObject()


def End(builder):
    return AuthScramWelcomeEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest

from autobahn.wamp.exception import InvalidUriError
from autobahn.wamp.message import (
    check_or_raise_realm_name,
    identify_realm_name_category,
)


class TestWampIdentifiers(unittest.TestCase):
    def test_valid_realm_names(self):
        for name in [
            "realm1",
            "com.example.myapp1",
            "myapp1.example.com",
            "eth.wamp-proto",
            "wamp-proto.eth",
            "eth.wamp-proto.myapp1",
            "myapp1.wamp-proto.eth",
            "aaa",
            "Abc",
            "a00",
            "A00",
            "0x0000000000000000000000000000000000000000",
            "0xe59C7418403CF1D973485B36660728a5f4A8fF9c",
        ]:
            self.assertEqual(name, check_or_raise_realm_name(name))

    def test_invalid_realm_names(self):
        for name in [
            None,
            23,
            {},
            "",
            ".realm1",
            "123realm",
            "0x" + "00" * 64,
            "0x" + "00" * 32,
            "0x" + "zz" * 40,
            "rlm$test",
            "a" * 256,
        ]:
            self.assertRaises(InvalidUriError, check_or_raise_realm_name, name)

    def test_realm_name_categories(self):
        for name, category in [
            # valid
            ("realm1", "standalone"),
            ("com.example.myapp1", "standalone"),
            ("myapp1.example.com", "standalone"),
            ("eth.wamp-proto", "reverse_ens"),
            ("wamp-proto.eth", "ens"),
            ("eth.wamp-proto.myapp1", "reverse_ens"),
            ("myapp1.wamp-proto.eth", "ens"),
            ("aaa", "standalone"),
            ("Abc", "standalone"),
            ("a00", "standalone"),
            ("A00", "standalone"),
            ("0x0000000000000000000000000000000000000000", "eth"),
            ("0xe59C7418403CF1D973485B36660728a5f4A8fF9c", "eth"),
            # invalid
            (None, None),
            (23, None),
            ({}, None),
            ("", None),
            (".realm1", None),
            ("123realm", None),
            ("0x" + "00" * 64, None),
            ("0x" + "00" * 32, None),
            ("0x" + "zz" * 40, None),
            ("rlm$test", None),
            ("a" * 256, None),
        ]:
            self.assertEqual(category, identify_realm_name_category(name))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class EventReceived(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = EventReceived()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsEventReceived(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # EventReceived
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # EventReceived
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # EventReceived
    def Publication(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # EventReceived
    def ForwardFor(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from autobahn.wamp.gen.wamp.proto.Principal import Principal

            obj = Principal()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # EventReceived
    def ForwardForLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # EventReceived
    def ForwardForIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0


def EventReceivedStart(builder):
    builder.StartObject(3)


def Start(builder):
    EventReceivedStart(builder)


def EventReceivedAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    EventReceivedAddSession(builder, session)


def EventReceivedAddPublication(builder, publication):
    builder.PrependUint64Slot(1, publication, 0)


def AddPublication(builder, publication):
    EventReceivedAddPublication(builder, publication)


def EventReceivedAddForwardFor(builder, forwardFor):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(forwardFor), 0
    )


def AddForwardFor(builder, forwardFor):
    EventReceivedAddForwardFor(builder, forwardFor)


def EventReceivedStartForwardForVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartForwardForVector(builder, numElems):
    return EventReceivedStartForwardForVector(builder, numElems)


def EventReceivedEnd(builder):
    return builder.EndObject()


def End(builder):
    return EventReceivedEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest
from binascii import a2b_hex

from autobahn.wamp.auth import derive_scram_credential

TEST_VECTORS = [
    {
        "email": "foobar@example.com",
        "password": "secret123",
        "salt": None,
        "expected": {
            "iterations": 4096,
            "kdf": "argon2id-13",
            "memory": 512,
            "salt": "3bc3ca01dd1d501ca1c22e1c5d7d16fe",
            "server-key": "8de7864c316f3c2356fd76cfdab696db55bc70e680fe5180e2f731e2345acca2",
            "stored-key": "e796c2f0a51770303ee4616bc630a66774d51a55003154aff2a54ec7c4ac0e38",
        },
    },
    {
        "email": "foobar@example.com",
        "password": "secret123",
        "salt": a2b_hex("ae1f0d2f422757809077785e660b62c6"),
        "expected": {
            "iterations": 4096,
            "kdf": "argon2id-13",
            "memory": 512,
            "salt": "ae1f0d2f422757809077785e660b62c6",
            "server-key": "0d8e7e9222a7c0e54c9e979aa342115699ff5696c45dc379b5ee241338a5861d",
            "stored-key": "5f19358ff6f38e267b6ef1ea1d862514ec4e8745a84682259fd3894be09febb5",
        },
    },
]


class TestKey(unittest.TestCase):
    def test_derive_scram_credential(self):
        for tv in TEST_VECTORS:
            email = tv["email"]
            password = tv["password"]
            salt = tv["salt"]
            expected = tv["expected"]
            credential = derive_scram_credential(email, password, salt)
            self.assertEqual(credential, expected)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import binascii
import hashlib
import os
import unittest
from unittest.mock import Mock

import txaio

if os.environ.get("USE_TWISTED", None):
    txaio.use_twisted()
elif os.environ.get("USE_ASYNCIO", None):
    txaio.use_asyncio()
else:
    raise RuntimeError("need either USE_TWISTED=1 or USE_ASYNCIO=1")

from autobahn.wamp import types
from autobahn.wamp.auth import create_authenticator
from autobahn.wamp.cryptosign import HAS_CRYPTOSIGN, CryptosignAuthextra, _makepad

if HAS_CRYPTOSIGN:
    from nacl.encoding import HexEncoder

    from autobahn.wamp.cryptosign import CryptosignKey


import tempfile

keybody = """-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAAAMwAAAAtzc2gtZW
QyNTUxOQAAACAa38i/4dNWFuZN/72QAJbyOwZvkUyML/u2b2B1uW4RbQAAAJj4FLyB+BS8
gQAAAAtzc2gtZWQyNTUxOQAAACAa38i/4dNWFuZN/72QAJbyOwZvkUyML/u2b2B1uW4RbQ
AAAEBNV9l6aPVVaWYgpthJwM5YJWhRjXKet1PcfHMt4oBFEBrfyL/h01YW5k3/vZAAlvI7
Bm+RTIwv+7ZvYHW5bhFtAAAAFXNvbWV1c2VyQGZ1bmt0aGF0LmNvbQ==
-----END OPENSSH PRIVATE KEY-----"""

pubkey = """ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIJVp3hjHwIQyEladzd8mFcf0YSXcmyKS3qMLB7VqTQKm someuser@example.com
"""

# valid test vectors for WAMP-cryptosign signature testing
test_vectors_1 = [
    # _WITHOUT_ channel_id
    {
        "channel_id": None,
        "private_key": "4d57d97a68f555696620a6d849c0ce582568518d729eb753dc7c732de2804510",
        "challenge": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
        "signature": "b32675b221f08593213737bef8240e7c15228b07028e19595294678c90d11c0cae80a357331bfc5cc9fb71081464e6e75013517c2cf067ad566a6b7b728e5d03ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
    },
    {
        "channel_id": None,
        "private_key": "d511fe78e23934b3dadb52fcd022974b80bd92bccc7c5cf404e46cc0a8a2f5cd",
        "challenge": "b26c1f87c13fc1da14997f1b5a71995dff8fbe0a62fae8473c7bdbd05bfb607d",
        "signature": "d4209ad10d5aff6bfbc009d7e924795de138a63515efc7afc6b01b7fe5201372190374886a70207b042294af5bd64ce725cd8dceb344e6d11c09d1aaaf4d660fb26c1f87c13fc1da14997f1b5a71995dff8fbe0a62fae8473c7bdbd05bfb607d",
    },
    {
        "channel_id": None,
        "private_key": "6e1fde9cf9e2359a87420b65a87dc0c66136e66945196ba2475990d8a0c3a25b",
        "challenge": "b05e6b8ad4d69abf74aa3be3c0ee40ae07d66e1895b9ab09285a2f1192d562d2",
        "signature": "7beb282184baadd08f166f16dd683b39cab53816ed81e6955def951cb2ddad1ec184e206746fd82bda075af03711d3d5658fc84a76196b0fa8d1ebc92ef9f30bb05e6b8ad4d69abf74aa3be3c0ee40ae07d66e1895b9ab09285a2f1192d562d2",
    },
    # _WITH_ channel_id
    {
        "channel_id": "62e935ae755f3d48f80d4d59f6121358c435722a67e859cc0caa8b539027f2ff",
        "private_key": "4d57d97a68f555696620a6d849c0ce582568518d729eb753dc7c732de2804510",
        "challenge": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
        "signature": "9b6f41540c9b95b4b7b281c3042fa9c54cef43c842d62ea3fd6030fcb66e70b3e80d49d44c29d1635da9348d02ec93f3ed1ef227dfb59a07b580095c2b82f80f9d16ca518aa0c2b707f2b2a609edeca73bca8dd59817a633f35574ac6fd80d00",
    },
    {
        "channel_id": "62e935ae755f3d48f80d4d59f6121358c435722a67e859cc0caa8b539027f2ff",
        "private_key": "d511fe78e23934b3dadb52fcd022974b80bd92bccc7c5cf404e46cc0a8a2f5cd",
        "challenge": "b26c1f87c13fc1da14997f1b5a71995dff8fbe0a62fae8473c7bdbd05bfb607d",
        "signature": "305aaa3ac25e98f651427688b3fc43fe7d8a68a7ec1d7d61c61517c519bd4a427c3015599d83ca28b4c652333920223844ef0725eb5dc2febfd6af7677b73f01d0852a29b460fc92ec943242ac638a053bbacc200512b18b30d15083cbdc9282",
    },
    {
        "channel_id": "62e935ae755f3d48f80d4d59f6121358c435722a67e859cc0caa8b539027f2ff",
        "private_key": "6e1fde9cf9e2359a87420b65a87dc0c66136e66945196ba2475990d8a0c3a25b",
        "challenge": "b05e6b8ad4d69abf74aa3be3c0ee40ae07d66e1895b9ab09285a2f1192d562d2",
        "signature": "ee3c7644fd8070532bc1fde3d70d742267da545d8c8f03e63bda63f1ad4214f4d2c4bfdb4eb9526def42deeb7e31602a6ff99eba893e0a4ad4d45892ca75e608d2b75e24a189a7f78ca776ba36fc53f6c3e31c32f251f2c524f0a44202f2902d",
    },
]


class TestSigVectors(unittest.TestCase):
    def test_vectors(self):
        session = Mock()

        for testvec in test_vectors_1:
            # setup fake transport details including fake channel_id
            if testvec["channel_id"]:
                channel_id = binascii.a2b_hex(testvec["channel_id"])
                channel_id_type = "tls-unique"
                session._transport.transport_details = types.TransportDetails(
                    channel_id={"tls-unique": channel_id}
                )
            else:
                channel_id = None
                channel_id_type = None
                session._transport.transport_details = types.TransportDetails(
                    channel_id=None
                )

            # private signing key (the seed for it)
            private_key = CryptosignKey.from_bytes(
                binascii.a2b_hex(testvec["private_key"])
            )

            # the fake challenge we've received
            challenge = types.Challenge(
                "cryptosign", dict(challenge=testvec["challenge"])
            )

            # ok, now sign the challenge
            f_signed = private_key.sign_challenge(
                challenge, channel_id=channel_id, channel_id_type=channel_id_type
            )

            def success(signed):
                # the signature returned is a Hex encoded string
                self.assertTrue(type(signed) == str)

                # we return the concatenation of the signature and the message signed (96 bytes)
                self.assertEqual(
                    192,
                    len(signed),
                )

                # must match the expected value in our test vector
                self.assertEqual(
                    testvec["signature"],
                    signed,
                )

            def failed(err):
                self.fail(str(err))

            txaio.add_callbacks(f_signed, success, failed)


class TestAuth(unittest.TestCase):
    def setUp(self):
        self.key = CryptosignKey.from_ssh_bytes(keybody)
        self.privkey_hex = self.key._key.encode(encoder=HexEncoder)

        # all tests here fake the use of channel_id_type='tls-unique' with the following channel_id
        m = hashlib.sha256()
        m.update(b"some TLS message")

        # 62e935ae755f3d48f80d4d59f6121358c435722a67e859cc0caa8b539027f2ff
        channel_id = m.digest()

        self.transport_details = types.TransportDetails(
            channel_id={"tls-unique": channel_id}
        )

    def test_public_key(self):
        self.assertEqual(
            self.key.public_key(binary=False),
            "1adfc8bfe1d35616e64dffbd900096f23b066f914c8c2ffbb66f6075b96e116d",
        )

    def test_valid(self):
        session = Mock()
        session._transport.transport_details = self.transport_details

        challenge = types.Challenge("cryptosign", dict(challenge="ff" * 32))
        f_signed = self.key.sign_challenge(
            challenge,
            channel_id=self.transport_details.channel_id["tls-unique"],
            channel_id_type="tls-unique",
        )

        def success(signed):
            self.assertEqual(
                192,
                len(signed),
            )
            self.assertEqual(
                "9b6f41540c9b95b4b7b281c3042fa9c54cef43c842d62ea3fd6030fcb66e70b3e80d49d44c29d1635da9348d02ec93f3ed1ef227dfb59a07b580095c2b82f80f9d16ca518aa0c2b707f2b2a609edeca73bca8dd59817a633f35574ac6fd80d00",
                signed,
            )

        def failed(err):
            self.fail(str(err))

        txaio.add_callbacks(f_signed, success, failed)

    def test_authenticator(self):
        authenticator = create_authenticator(
            "cryptosign",
            authid="someone",
            authextra={"channel_binding": "tls-unique"},
            privkey=self.privkey_hex,
        )
        session = Mock()
        session._transport.transport_details = self.transport_details
        challenge = types.Challenge("cryptosign", dict(challenge="ff" * 32))
        f_reply = authenticator.on_challenge(session, challenge)

        def success(reply):
            self.assertEqual(
                reply,
                "9b6f41540c9b95b4b7b281c3042fa9c54cef43c842d62ea3fd6030fcb66e70b3e80d49d44c29d1635da9348d02ec93f3ed1ef227dfb59a07b580095c2b82f80f9d16ca518aa0c2b707f2b2a609edeca73bca8dd59817a633f35574ac6fd80d00",
            )

        def failed(err):
            self.fail(str(err))

        txaio.add_callbacks(f_reply, success, failed)


class TestKey(unittest.TestCase):
    def test_pad(self):
        self.assertEqual(_makepad(0), b"")
        self.assertEqual(_makepad(2), b"\x01\x02")
        self.assertEqual(_makepad(3), b"\x01\x02\x03")
        self.assertEqual(
            _makepad(30),
            b"\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e",
        )
        self.assertEqual(
            binascii.b2a_hex(_makepad(30)).decode(),
            "0102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e",
        )

    def test_key(self):
        with tempfile.NamedTemporaryFile("w+t") as fp:
            fp.write(keybody)
            fp.seek(0)

            key = CryptosignKey.from_ssh_file(fp.name)
            self.assertEqual(
                key.public_key(),
                "1adfc8bfe1d35616e64dffbd900096f23b066f914c8c2ffbb66f6075b96e116d",
            )

    def test_pubkey(self):
        with tempfile.NamedTemporaryFile("w+t") as fp:
            fp.write(pubkey)
            fp.seek(0)

            key = CryptosignKey.from_ssh_file(fp.name)
            self.assertEqual(
                key.public_key(binary=False),
                "9569de18c7c0843212569dcddf2615c7f46125dc9b2292dea30b07b56a4d02a6",
            )
            self.assertEqual(key.comment, "someuser@example.com")


class TestAuthExtra(unittest.TestCase):
    def test_default_ctor(self):
        ae = CryptosignAuthextra()
        self.assertEqual(ae.marshal(), {})

    def test_ctor(self):
        ae1 = CryptosignAuthextra(pubkey=b"\xff" * 32)
        self.assertEqual(ae1.marshal(), {"pubkey": "ff" * 32})

        ae1 = CryptosignAuthextra(pubkey=b"\xff" * 32, bandwidth=200)
        self.assertEqual(
            ae1.marshal(), {"pubkey": "ff" * 32, "reservation": {"bandwidth": 200}}
        )

    def test_parse(self):
        data_original = {
            "pubkey": "9019a424b040859c108edee02e64c1dcb32b253686d7b5db56c306e9bdb2fe7e",
            "challenge": "fe81c84e94a75a357c259d6b37361e43966a45f57dff181bb61b2f91a0f4ac88",
            "channel_binding": "tls-unique",
            "channel_id": "2e642bf991f48ece9133a0a32d15550921dda12bfebfbc941571d4b2960540bc",
            "trustroot": "0xe78ea2fE1533D4beD9A10d91934e109A130D0ad8",
            "reservation": {
                "chain_id": 999,
                "block_no": 123456789,
                "realm": "0x163D58cE482560B7826b4612f40aa2A7d53310C4",
                "delegate": "0x72b3486d38E9f49215b487CeAaDF27D6acf22115",
                "seeder": "0x52d66f36A7927cF9612e1b40bD6549d08E0513Ff",
                "bandwidth": 200,
            },
            "signature": "747763c69394270603f64af5be3f8256a14b41ff51027e583ee81db9f1f15a01cc8e55218a76139f26dbaaa78d8a537d80d248b3fc6245ecf4602cc5fbb0f6452e",
        }
        ae1 = CryptosignAuthextra.parse(data_original)
        data_marshalled = ae1.marshal()

        # FIXME: marshal check-summed eth addresses
        data_original["trustroot"] = data_original["trustroot"].lower()
        for k in ["realm", "delegate", "seeder"]:
            data_original["reservation"][k] = data_original["reservation"][k].lower()

        self.assertEqual(data_marshalled, data_original)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# Subscription topic matching method.
class Match(object):
    # Match URI exact.
    EXACT = 0
    # Match URI by prefix.
    PREFIX = 1
    # Match URI by wildcard.
    WILDCARD = 2



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os

if os.environ.get("USE_TWISTED", False):
    from autobahn.twisted.wamp import ApplicationSession
    from autobahn.wamp import message, role
    from autobahn.wamp.exception import ProtocolError
    from autobahn.wamp.types import TransportDetails
    from twisted.internet import defer
    from twisted.python.failure import Failure
    from twisted.trial import unittest

    class MockTransport:
        def __init__(self):
            self.messages = []
            self._transport_details = TransportDetails()

        def transport_details(self):
            return self._transport_details

        def send(self, msg):
            self.messages.append(msg)

        def close(self, *args, **kw):
            pass

    class MockApplicationSession(ApplicationSession):
        """
        This is used by tests, which typically attach their own handler to
        on*() methods. This just collects any errors from onUserError
        """

        def __init__(self, *args, **kw):
            ApplicationSession.__init__(self, *args, **kw)
            self.errors = []
            self._realm = "dummy"
            self._transport = MockTransport()

        def onUserError(self, e, msg):
            self.errors.append((e.value, msg))

    def exception_raiser(exc):
        """
        Create a method that takes any args and always raises the given
        Exception instance.
        """
        assert isinstance(exc, Exception), "Must derive from Exception"

        def method(*args, **kw):
            raise exc

        return method

    def async_exception_raiser(exc):
        """
        Create a method that takes any args, and always returns a Deferred
        that has failed.
        """
        assert isinstance(exc, Exception), "Must derive from Exception"

        def method(*args, **kw):
            try:
                raise exc
            except:
                return defer.fail(Failure())

        return method

    def create_mock_welcome():
        return message.Welcome(
            1234,
            {
                "broker": role.RoleBrokerFeatures(),
            },
        )

    class TestSessionCallbacks(unittest.TestCase):
        """
        These test that callbacks on user-overridden ApplicationSession
        methods that produce errors are handled correctly.

        XXX should do state-diagram documenting where we are when each
        of these cases arises :/
        """

        # XXX sure would be nice to use py.test @fixture to do the
        # async/sync exception-raising stuff (i.e. make each test run
        # twice)...but that would mean switching all test-running over
        # to py-test

        # the whole variable must not be defined to deactivate (!)
        skip = True

        def test_on_join(self):
            session = MockApplicationSession()
            exception = RuntimeError("blammo")
            session.onJoin = exception_raiser(exception)
            msg = create_mock_welcome()

            # give the sesion a WELCOME, from which it should call onJoin
            session.onMessage(msg)

            # make sure we got the right error out of onUserError
            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_join_deferred(self):
            session = MockApplicationSession()
            exception = RuntimeError("blammo")
            session.onJoin = async_exception_raiser(exception)
            msg = create_mock_welcome()

            # give the sesion a WELCOME, from which it should call onJoin
            session.onMessage(msg)

            # make sure we got the right error out of onUserError
            # import traceback
            # traceback.print_exception(*session.errors[0][:3])
            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_leave(self):
            session = MockApplicationSession()
            exception = RuntimeError("boom")
            session.onLeave = exception_raiser(exception)
            msg = message.Abort("testing")

            # we haven't done anything, so this is "abort before we've
            # connected"
            session.onMessage(msg)

            # make sure we got the right error out of onUserError
            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_leave_deferred(self):
            session = MockApplicationSession()
            exception = RuntimeError("boom")
            session.onLeave = async_exception_raiser(exception)
            msg = message.Abort("testing")

            # we haven't done anything, so this is "abort before we've
            # connected"
            session.onMessage(msg)

            # make sure we got the right error out of onUserError
            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_leave_valid_session(self):
            """
            cover when onLeave called after we have a valid session
            """
            session = MockApplicationSession()
            exception = RuntimeError("such challenge")
            session.onLeave = exception_raiser(exception)
            # we have to get to an established connection first...
            session.onMessage(create_mock_welcome())
            self.assertTrue(session._session_id is not None)

            # okay we have a session ("because ._session_id is not None")
            msg = message.Goodbye()
            session.onMessage(msg)

            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_leave_valid_session_deferred(self):
            """
            cover when onLeave called after we have a valid session
            """
            session = MockApplicationSession()
            exception = RuntimeError("such challenge")
            session.onLeave = async_exception_raiser(exception)
            # we have to get to an established connection first...
            session.onMessage(create_mock_welcome())
            self.assertTrue(session._session_id is not None)

            # okay we have a session ("because ._session_id is not None")
            msg = message.Goodbye()
            session.onMessage(msg)

            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_leave_after_bad_challenge(self):
            """
            onLeave raises error after onChallenge fails
            """
            session = MockApplicationSession()
            exception = RuntimeError("such challenge")
            session.onLeave = exception_raiser(exception)
            session.onChallenge = exception_raiser(exception)
            # make a challenge (which will fail, and then the
            # subsequent onLeave will also fail)
            msg = message.Challenge("foo")
            session.onMessage(msg)

            self.assertEqual(2, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_disconnect_via_close(self):
            session = MockApplicationSession()
            exception = RuntimeError("sideways")
            session.onDisconnect = exception_raiser(exception)
            # we short-cut the whole state-machine traversal here by
            # just calling onClose directly, which would normally be
            # called via a Protocol, e.g.,
            # autobahn.wamp.websocket.WampWebSocketProtocol
            session.onClose(False)

            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_disconnect_via_close_deferred(self):
            session = MockApplicationSession()
            exception = RuntimeError("sideways")
            session.onDisconnect = async_exception_raiser(exception)
            # we short-cut the whole state-machine traversal here by
            # just calling onClose directly, which would normally be
            # called via a Protocol, e.g.,
            # autobahn.wamp.websocket.WampWebSocketProtocol
            session.onClose(False)

            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        # XXX FIXME Probably more ways to call onLeave!

        def test_on_challenge(self):
            session = MockApplicationSession()
            exception = RuntimeError("such challenge")
            session.onChallenge = exception_raiser(exception)
            msg = message.Challenge("foo")

            # execute
            session.onMessage(msg)

            # we already handle any onChallenge errors as "abort the
            # connection". So make sure our error showed up in the
            # fake-transport.
            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])
            self.assertEqual(1, len(session._transport.messages))
            reply = session._transport.messages[0]
            self.assertIsInstance(reply, message.Abort)
            self.assertEqual("such challenge", reply.message)

        def test_on_challenge_deferred(self):
            session = MockApplicationSession()
            exception = RuntimeError("such challenge")
            session.onChallenge = async_exception_raiser(exception)
            msg = message.Challenge("foo")

            # execute
            session.onMessage(msg)

            # we already handle any onChallenge errors as "abort the
            # connection". So make sure our error showed up in the
            # fake-transport.
            self.assertEqual(1, len(session.errors))
            self.assertEqual(session.errors[0][0], exception)
            self.assertEqual(1, len(session._transport.messages))
            reply = session._transport.messages[0]
            self.assertIsInstance(reply, message.Abort)
            self.assertEqual("such challenge", reply.message)

        def test_no_session(self):
            """
            test "all other cases" when we don't yet have a session
            established, which should all raise ProtocolErrors and
            *not* go through the onUserError handler. We cheat and
            just test one.
            """
            session = MockApplicationSession()
            exception = RuntimeError("such challenge")
            session.onConnect = exception_raiser(exception)

            for msg in [message.Goodbye()]:
                self.assertRaises(ProtocolError, session.onMessage, (msg,))
            self.assertEqual(0, len(session.errors))

        def test_on_disconnect(self):
            session = MockApplicationSession()
            exception = RuntimeError("oh sadness")
            session.onDisconnect = exception_raiser(exception)
            # we short-cut the whole state-machine traversal here by
            # just calling onClose directly, which would normally be
            # called via a Protocol, e.g.,
            # autobahn.wamp.websocket.WampWebSocketProtocol
            session.onClose(False)

            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_disconnect_deferred(self):
            session = MockApplicationSession()
            exception = RuntimeError("oh sadness")
            session.onDisconnect = async_exception_raiser(exception)
            # we short-cut the whole state-machine traversal here by
            # just calling onClose directly, which would normally be
            # called via a Protocol, e.g.,
            # autobahn.wamp.websocket.WampWebSocketProtocol
            session.onClose(False)

            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_disconnect_with_session(self):
            session = MockApplicationSession()
            exception = RuntimeError("the pain runs deep")
            session.onDisconnect = exception_raiser(exception)
            # create a valid session
            session.onMessage(create_mock_welcome())

            # we short-cut the whole state-machine traversal here by
            # just calling onClose directly, which would normally be
            # called via a Protocol, e.g.,
            # autobahn.wamp.websocket.WampWebSocketProtocol
            session.onClose(False)

            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_disconnect_with_session_deferred(self):
            session = MockApplicationSession()
            exception = RuntimeError("the pain runs deep")
            session.onDisconnect = async_exception_raiser(exception)
            # create a valid session
            session.onMessage(create_mock_welcome())

            # we short-cut the whole state-machine traversal here by
            # just calling onClose directly, which would normally be
            # called via a Protocol, e.g.,
            # autobahn.wamp.websocket.WampWebSocketProtocol
            session.onClose(False)

            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_connect(self):
            session = MockApplicationSession()
            exception = RuntimeError("the pain runs deep")
            session.onConnect = exception_raiser(exception)
            trans = MockTransport()

            # normally would be called from a Protocol?
            session.onOpen(trans)

            # shouldn't have done the .join()
            self.assertEqual(0, len(trans.messages))
            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        def test_on_connect_deferred(self):
            session = MockApplicationSession()
            exception = RuntimeError("the pain runs deep")
            session.onConnect = async_exception_raiser(exception)
            trans = MockTransport()

            # normally would be called from a Protocol?
            session.onOpen(trans)

            # shouldn't have done the .join()
            self.assertEqual(0, len(trans.messages))
            self.assertEqual(1, len(session.errors))
            self.assertEqual(exception, session.errors[0][0])

        # XXX likely missing other ways to invoke the above. need to
        # cover, for sure:
        #
        # onChallenge
        # onJoin
        # onLeave
        # onDisconnect
        #
        # what about other ISession ones?
        # onConnect
        # onDisconnect

        # NOTE: for Event stuff, that is publish() handlers,
        # test_publish_callback_exception in test_protocol.py already
        # covers exceptions coming from user-code.



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# Import smoke test for the framework-agnostic core modules.
#
# This guards against import-time regressions that only surface on some Python
# versions - in particular eager evaluation of annotations on CPython < 3.14
# (PEP 649 defers it on 3.14): a string forward-reference combined with
# ``| None`` (e.g. ``"ISecurityModule" | None``) raises ``TypeError`` at
# class-definition time. See issue #1878.
#
# We import an explicit, curated set rather than walking every submodule: a full
# single-process walk is unreliable for autobahn because (a) Twisted and asyncio
# bindings are mutually exclusive in one process (txaio backend selection) and
# (b) several modules require optional extras (snappy, flatbuffers runtime).
# These core modules are framework-agnostic and must always import with the
# crypto extras installed.
#
# IMPORTANT: run with crypto extras (``autobahn[all]`` / ``[encryption]`` ->
# PyNaCl) so the ``autobahn.wamp.cryptosign`` ``HAS_CRYPTOSIGN`` code paths
# (where #1878 lived) are actually exercised - see ``test_crypto_extras_present``.

from __future__ import annotations

import importlib

import pytest

CORE_MODULES = [
    "autobahn",
    "autobahn.util",
    "autobahn.wamp",
    "autobahn.wamp.cryptosign",  # regressed in 26.6.1 -> see #1878
    "autobahn.wamp.auth",
    "autobahn.wamp.interfaces",
    "autobahn.wamp.types",
    "autobahn.wamp.message",
    "autobahn.wamp.role",
    "autobahn.wamp.serializer",
    "autobahn.wamp.component",
    "autobahn.websocket.protocol",
    "autobahn.websocket.types",
    "autobahn.websocket.compress",
]


@pytest.mark.parametrize("modname", CORE_MODULES)
def test_import(modname):
    """
    Importing each core autobahn module must not raise.
    """
    importlib.import_module(modname)


def test_crypto_extras_present():
    """
    The cryptosign import guard is only meaningful when PyNaCl is installed,
    since the ``autobahn.wamp.cryptosign`` class bodies (and thus the #1878
    regression) are gated behind ``HAS_CRYPTOSIGN``.
    """
    import autobahn.wamp.cryptosign as cryptosign

    assert (
        cryptosign.HAS_CRYPTOSIGN
    ), "install crypto extras (autobahn[all] / [encryption]) so this import guard is exercised"



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class DealerFeatures(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DealerFeatures()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsDealerFeatures(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # DealerFeatures
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DealerFeatures
    def CallerIdentification(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def CallTrustlevels(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def CallTimeout(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def CallCanceling(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def ProgressiveCallResults(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def RegistrationRevocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def PatternBasedRegistration(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def SharedRegistration(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def SessionMetaApi(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def RegistrationMetaApi(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def TestamentMetaApi(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def PayloadTransparency(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # DealerFeatures
    def PayloadEncryptionCryptobox(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False


def DealerFeaturesStart(builder):
    builder.StartObject(13)


def Start(builder):
    DealerFeaturesStart(builder)


def DealerFeaturesAddCallerIdentification(builder, callerIdentification):
    builder.PrependBoolSlot(0, callerIdentification, 0)


def AddCallerIdentification(builder, callerIdentification):
    DealerFeaturesAddCallerIdentification(builder, callerIdentification)


def DealerFeaturesAddCallTrustlevels(builder, callTrustlevels):
    builder.PrependBoolSlot(1, callTrustlevels, 0)


def AddCallTrustlevels(builder, callTrustlevels):
    DealerFeaturesAddCallTrustlevels(builder, callTrustlevels)


def DealerFeaturesAddCallTimeout(builder, callTimeout):
    builder.PrependBoolSlot(2, callTimeout, 0)


def AddCallTimeout(builder, callTimeout):
    DealerFeaturesAddCallTimeout(builder, callTimeout)


def DealerFeaturesAddCallCanceling(builder, callCanceling):
    builder.PrependBoolSlot(3, callCanceling, 0)


def AddCallCanceling(builder, callCanceling):
    DealerFeaturesAddCallCanceling(builder, callCanceling)


def DealerFeaturesAddProgressiveCallResults(builder, progressiveCallResults):
    builder.PrependBoolSlot(4, progressiveCallResults, 0)


def AddProgressiveCallResults(builder, progressiveCallResults):
    DealerFeaturesAddProgressiveCallResults(builder, progressiveCallResults)


def DealerFeaturesAddRegistrationRevocation(builder, registrationRevocation):
    builder.PrependBoolSlot(5, registrationRevocation, 0)


def AddRegistrationRevocation(builder, registrationRevocation):
    DealerFeaturesAddRegistrationRevocation(builder, registrationRevocation)


def DealerFeaturesAddPatternBasedRegistration(builder, patternBasedRegistration):
    builder.PrependBoolSlot(6, patternBasedRegistration, 0)


def AddPatternBasedRegistration(builder, patternBasedRegistration):
    DealerFeaturesAddPatternBasedRegistration(builder, patternBasedRegistration)


def DealerFeaturesAddSharedRegistration(builder, sharedRegistration):
    builder.PrependBoolSlot(7, sharedRegistration, 0)


def AddSharedRegistration(builder, sharedRegistration):
    DealerFeaturesAddSharedRegistration(builder, sharedRegistration)


def DealerFeaturesAddSessionMetaApi(builder, sessionMetaApi):
    builder.PrependBoolSlot(8, sessionMetaApi, 0)


def AddSessionMetaApi(builder, sessionMetaApi):
    DealerFeaturesAddSessionMetaApi(builder, sessionMetaApi)


def DealerFeaturesAddRegistrationMetaApi(builder, registrationMetaApi):
    builder.PrependBoolSlot(9, registrationMetaApi, 0)


def AddRegistrationMetaApi(builder, registrationMetaApi):
    DealerFeaturesAddRegistrationMetaApi(builder, registrationMetaApi)


def DealerFeaturesAddTestamentMetaApi(builder, testamentMetaApi):
    builder.PrependBoolSlot(10, testamentMetaApi, 0)


def AddTestamentMetaApi(builder, testamentMetaApi):
    DealerFeaturesAddTestamentMetaApi(builder, testamentMetaApi)


def DealerFeaturesAddPayloadTransparency(builder, payloadTransparency):
    builder.PrependBoolSlot(11, payloadTransparency, 0)


def AddPayloadTransparency(builder, payloadTransparency):
    DealerFeaturesAddPayloadTransparency(builder, payloadTransparency)


def DealerFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    builder.PrependBoolSlot(12, payloadEncryptionCryptobox, 0)


def AddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    DealerFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox)


def DealerFeaturesEnd(builder):
    return builder.EndObject()


def End(builder):
    return DealerFeaturesEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from autobahn._version import __version__

version = __version__

import os

import txaio

from autobahn import flatbuffers


def check_zlmdb_flatbuffers_version_in_sync() -> str:
    """
    Check that autobahn and zlmdb have the same vendored flatbuffers version.

    This is important for applications like Crossbar.io that use both autobahn
    (for data-in-transit) and zlmdb (for data-at-rest) with FlatBuffers
    serialization. When sending a FlatBuffers database record as a WAMP
    application payload, both libraries must use compatible FlatBuffers
    runtimes to avoid subtle serialization issues.

    :returns: The vendored flatbuffers version string (e.g. "25.12.19")
              if both are in sync.
    :raises RuntimeError: If the versions differ.
    :raises ImportError: If zlmdb is not installed.

    Example::

        import autobahn
        version = autobahn.check_zlmdb_flatbuffers_version_in_sync()
        print(f"FlatBuffers version: {version}")
    """
    import zlmdb.flatbuffers

    autobahn_version = flatbuffers.__version__
    zlmdb_version = zlmdb.flatbuffers.__version__

    if autobahn_version != zlmdb_version:
        raise RuntimeError(
            f"FlatBuffers version mismatch: autobahn has {autobahn_version!r}, "
            f"zlmdb has {zlmdb_version!r}. Both should be the same for "
            f"reliable data-in-transit/data-at-rest interoperability."
        )

    return autobahn_version

# this is used in the unit tests (trial/pytest), and when already done here, there
# is no risk and headaches with finding out if/where an import implies a framework
if os.environ.get("USE_TWISTED", False) and os.environ.get("USE_ASYNCIO", False):
    raise RuntimeError("fatal: _both_ USE_TWISTED and USE_ASYNCIO are set!")

if os.environ.get("USE_TWISTED", False):
    txaio.use_twisted()
elif os.environ.get("USE_ASYNCIO", False):
    txaio.use_asyncio()
else:
    # neither USE_TWISTED nor USE_ASYNCIO selected from env var
    pass



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import encode
from . import number_types as N


class Table(object):
  """Table wraps a byte slice and provides read access to its data.

  The variable `Pos` indicates the root of the FlatBuffers object therein.
  """

  __slots__ = ("Bytes", "Pos")

  def __init__(self, buf, pos):
    N.enforce_number(pos, N.UOffsetTFlags)

    self.Bytes = buf
    self.Pos = pos

  def Offset(self, vtableOffset):
    """Offset provides access into the Table's vtable.

    Deprecated fields are ignored by checking the vtable's length.
    """

    vtable = self.Pos - self.Get(N.SOffsetTFlags, self.Pos)
    vtableEnd = self.Get(N.VOffsetTFlags, vtable)
    if vtableOffset < vtableEnd:
      return self.Get(N.VOffsetTFlags, vtable + vtableOffset)
    return 0

  def Indirect(self, off):
    """Indirect retrieves the relative offset stored at `offset`."""
    N.enforce_number(off, N.UOffsetTFlags)
    return off + encode.Get(N.UOffsetTFlags.packer_type, self.Bytes, off)

  def String(self, off):
    """String gets a string from data stored inside the flatbuffer."""
    N.enforce_number(off, N.UOffsetTFlags)
    off += encode.Get(N.UOffsetTFlags.packer_type, self.Bytes, off)
    start = off + N.UOffsetTFlags.bytewidth
    length = encode.Get(N.UOffsetTFlags.packer_type, self.Bytes, off)
    return bytes(self.Bytes[start : start + length])

  def VectorLen(self, off):
    """VectorLen retrieves the length of the vector whose offset is stored

    at "off" in this object.
    """
    N.enforce_number(off, N.UOffsetTFlags)

    off += self.Pos
    off += encode.Get(N.UOffsetTFlags.packer_type, self.Bytes, off)
    ret = encode.Get(N.UOffsetTFlags.packer_type, self.Bytes, off)
    return ret

  def Vector(self, off):
    """Vector retrieves the start of data of the vector whose offset is

    stored at "off" in this object.
    """
    N.enforce_number(off, N.UOffsetTFlags)

    off += self.Pos
    x = off + self.Get(N.UOffsetTFlags, off)
    # data starts after metadata containing the vector length
    x += N.UOffsetTFlags.bytewidth
    return x

  def Union(self, t2, off):
    """Union initializes any Table-derived type to point to the union at

    the given offset.
    """
    assert type(t2) is Table
    N.enforce_number(off, N.UOffsetTFlags)

    off += self.Pos
    t2.Pos = off + self.Get(N.UOffsetTFlags, off)
    t2.Bytes = self.Bytes

  def Get(self, flags, off):
    """Get retrieves a value of the type specified by `flags`  at the

    given offset.
    """
    N.enforce_number(off, N.UOffsetTFlags)
    return flags.py_type(encode.Get(flags.packer_type, self.Bytes, off))

  def GetSlot(self, slot, d, validator_flags):
    N.enforce_number(slot, N.VOffsetTFlags)
    if validator_flags is not None:
      N.enforce_number(d, validator_flags)
    off = self.Offset(slot)
    if off == 0:
      return d
    return self.Get(validator_flags, self.Pos + off)

  def GetVectorAsNumpy(self, flags, off):
    """GetVectorAsNumpy returns the vector that starts at `Vector(off)`

    as a numpy array with the type specified by `flags`. The array is
    a `view` into Bytes, so modifying the returned array will
    modify Bytes in place.
    """
    offset = self.Vector(off)
    length = self.VectorLen(off)  # TODO: length accounts for bytewidth, right?
    numpy_dtype = N.to_numpy_type(flags)
    return encode.GetVectorAsNumpy(numpy_dtype, self.Bytes, length, offset)

  def GetArrayAsNumpy(self, flags, off, length):
    """GetArrayAsNumpy returns the array with fixed width that starts at `Vector(offset)`

    with length `length` as a numpy array with the type specified by `flags`.
    The
    array is a `view` into Bytes so modifying the returned will modify Bytes in
    place.
    """
    numpy_dtype = N.to_numpy_type(flags)
    return encode.GetVectorAsNumpy(numpy_dtype, self.Bytes, length, off)

  def GetVOffsetTSlot(self, slot, d):
    """GetVOffsetTSlot retrieves the VOffsetT that the given vtable location

    points to. If the vtable value is zero, the default value `d`
    will be returned.
    """

    N.enforce_number(slot, N.VOffsetTFlags)
    N.enforce_number(d, N.VOffsetTFlags)

    off = self.Offset(slot)
    if off == 0:
      return d
    return off



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP transport channel framing (`channel_framing`).
class TransportChannelFraming(object):
    # Not set.
    NULL = 0
    # Raw transport itself is inherently message-based already (e.g. FUNCTION or VIRTIO).
    NATIVE = 1
    # Raw transport itself (e.g. TCP) is stream-based and channel framing applied is WebSocket (RFC6455).
    WEBSOCKET = 2
    # Raw transport itself (e.g. TCP) is stream-based and channel framing applied is RawSocket (WAMP).
    RAWSOCKET = 3



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import hashlib
import os
import threading
from typing import Any

from autobahn.wamp.types import TransportDetails
from twisted.internet.address import IPv4Address, UNIXAddress
from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IProcessTransport, ITransport

try:
    from twisted.internet.stdio import PipeAddress
except ImportError:
    # stdio.PipeAddress is only avail on Twisted 13.0+
    PipeAddress = type(None)

try:
    from twisted.internet.address import IPv6Address

    _HAS_IPV6 = True
except ImportError:
    _HAS_IPV6 = False
    IPv6Address = type(None)

try:
    from OpenSSL.SSL import Connection

    from twisted.internet.interfaces import ISSLTransport
    from twisted.protocols.tls import TLSMemoryBIOProtocol

    _HAS_TLS = True
except ImportError:
    _HAS_TLS = False

__all = (
    "sleep",
    "peer2str",
    "transport_channel_id",
    "extract_peer_certificate",
    "create_transport_details",
)


def sleep(delay, reactor=None):
    """
    Inline sleep for use in co-routines (Twisted ``inlineCallback`` decorated functions).

    .. seealso::
       * `twisted.internet.defer.inlineCallbacks <http://twistedmatrix.com/documents/current/api/twisted.internet.defer.html#inlineCallbacks>`__
       * `twisted.internet.interfaces.IReactorTime <http://twistedmatrix.com/documents/current/api/twisted.internet.interfaces.IReactorTime.html>`__

    :param delay: Time to sleep in seconds.
    :type delay: float
    :param reactor: The Twisted reactor to use.
    :type reactor: None or provider of ``IReactorTime``.
    """
    if not reactor:
        from twisted.internet import reactor
    d = Deferred()
    reactor.callLater(delay, d.callback, None)
    return d


def peer2str(transport: ITransport | IProcessTransport) -> str:
    """
    Return a *peer descriptor* given a Twisted transport, for example:

    * ``tcp4:127.0.0.1:52914``: a TCPv4 socket
    * ``unix:/tmp/server.sock``: a Unix domain socket
    * ``process:142092``: a Pipe originating from a spawning (parent) process
    * ``pipe``: a Pipe terminating in a spawned (child) process

    :returns: Returns a string representation of the peer of the Twisted transport.
    """
    # IMPORTANT: we need to _first_ test for IProcessTransport
    if IProcessTransport.providedBy(transport):
        # note the PID of the forked process in the peer descriptor
        res = f"process:{transport.pid}"
    elif ITransport.providedBy(transport):
        addr: IPv4Address | IPv6Address | UNIXAddress | PipeAddress = (
            transport.getPeer()
        )
        if isinstance(addr, IPv4Address):
            res = f"tcp4:{addr.host}:{addr.port}"
        elif _HAS_IPV6 and isinstance(addr, IPv6Address):
            res = f"tcp6:{addr.host}:{addr.port}"
        elif isinstance(addr, UNIXAddress):
            if addr.name:
                res = f"unix:{addr.name}"
            else:
                res = "unix"
        elif isinstance(addr, PipeAddress):
            # sadly, we don't have a way to get at the PID of the other side of the pipe
            # res = "pipe"
            res = f"process:{os.getppid()}"
        else:
            # gracefully fallback if we can't map the peer's address
            res = "unknown"
    else:
        # gracefully fallback if we can't map the peer's transport
        res = "unknown"
    return res


if not _HAS_TLS:

    def transport_channel_id(
        transport: object, is_server: bool, channel_id_type: str | None = None
    ) -> bytes | None:
        if channel_id_type is None:
            return b"\x00" * 32
        else:
            raise RuntimeError(
                f'cannot determine TLS channel ID of type "{channel_id_type}" when TLS is not available on this system'
            )

else:

    def transport_channel_id(
        transport: object, is_server: bool, channel_id_type: str | None = None
    ) -> bytes | None:
        """
        Return TLS channel ID of WAMP transport of the given TLS channel ID type.

        Application-layer user authentication protocols are vulnerable to generic credential forwarding attacks,
        where an authentication credential sent by a client C to a server M may then be used by M to impersonate C at
        another server S.
        To prevent such credential forwarding attacks, modern authentication protocols rely on channel bindings.
        For example, WAMP-cryptosign can use the tls-unique channel identifier provided by the TLS layer to strongly
        bind authentication credentials to the underlying channel, so that a credential received on one TLS channel
        cannot be forwarded on another.

        :param transport: The Twisted TLS transport to extract the TLS channel ID from. If the transport isn't
            TLS based, and non-empty ``channel_id_type`` is requested, ``None`` will be returned. If the transport
            is indeed TLS based, an empty ``channel_id_type`` of ``None`` is requested, 32 NUL bytes will be returned.
        :param is_server: Flag indicating that the transport is a server transport.
        :param channel_id_type: TLS channel ID type, if set currently only ``"tls-unique"`` is supported.
        :returns: The TLS channel ID (32 bytes).
        """
        if channel_id_type is None:
            return b"\x00" * 32

        if channel_id_type not in ["tls-unique"]:
            raise RuntimeError(
                f'invalid TLS channel ID type "{channel_id_type}" requested'
            )

        if not isinstance(transport, TLSMemoryBIOProtocol):
            raise RuntimeError(
                f'cannot determine TLS channel ID of type "{channel_id_type}" when TLS is not available on this transport {type(transport)}'
            )

        # get access to the OpenSSL connection underlying the Twisted protocol
        # https://twistedmatrix.com/documents/current/api/twisted.protocols.tls.TLSMemoryBIOProtocol.html#getHandle
        connection: Connection = transport.getHandle()
        assert connection and isinstance(connection, Connection)

        # Obtain latest TLS Finished message that we expected from peer, or None if handshake is not completed.
        # http://www.pyopenssl.org/en/stable/api/ssl.html#OpenSSL.SSL.Connection.get_peer_finished
        is_not_resumed = True

        if channel_id_type == "tls-unique":
            # see also: https://bugs.python.org/file22646/tls_channel_binding.patch
            if is_server != is_not_resumed:
                # for routers (=servers) XOR new sessions, the channel ID is based on the TLS Finished message we
                # expected to receive from the client: contents of the message or None if the TLS handshake has
                # not yet completed.
                tls_finished_msg = connection.get_peer_finished()
            else:
                # for clients XOR resumed sessions, the channel ID is based on the TLS Finished message we sent
                # to the router (=server): contents of the message or None if the TLS handshake has not yet completed.
                tls_finished_msg = connection.get_finished()

            if tls_finished_msg is None:
                # this can occur when:
                #   1. we made a successful connection (in a TCP sense) but something failed with
                #      the TLS handshake (e.g. invalid certificate)
                #   2. the TLS handshake has not yet completed
                return b"\x00" * 32
            else:
                m = hashlib.sha256()
                m.update(tls_finished_msg)
                return m.digest()
        else:
            raise NotImplementedError(
                f'should not arrive here (unhandled channel_id_type "{channel_id_type}")'
            )


if not _HAS_TLS:

    def extract_peer_certificate(transport: object) -> dict[str, Any] | None:
        """
        Dummy when no TLS is available.

        :param transport: Ignored.
        :return: Always return ``None``.
        """
        return None

else:

    def extract_peer_certificate(
        transport: TLSMemoryBIOProtocol,
    ) -> dict[str, Any] | None:
        """
        Extract TLS x509 client certificate information from a Twisted stream transport, and
        return a dict with x509 TLS client certificate information (if the client provided a
        TLS client certificate).

        :param transport: The secure transport from which to extract the peer certificate (if present).
        :returns: If the peer provided a certificate, the parsed certificate information set.
        """
        # check if the Twisted transport is a TLSMemoryBIOProtocol
        if not (
            ISSLTransport.providedBy(transport)
            and hasattr(transport, "getPeerCertificate")
        ):
            return None

        cert = transport.getPeerCertificate()
        if cert:
            # extract x509 name components from an OpenSSL X509Name object
            def maybe_bytes(_value):
                if isinstance(_value, bytes):
                    return _value.decode("utf8")
                else:
                    return _value

            result = {
                "md5": "{}".format(maybe_bytes(cert.digest("md5"))).upper(),
                "sha1": "{}".format(maybe_bytes(cert.digest("sha1"))).upper(),
                "sha256": "{}".format(maybe_bytes(cert.digest("sha256"))).upper(),
                "expired": bool(cert.has_expired()),
                "hash": maybe_bytes(cert.subject_name_hash()),
                "serial": int(cert.get_serial_number()),
                "signature_algorithm": maybe_bytes(cert.get_signature_algorithm()),
                "version": int(cert.get_version()),
                "not_before": maybe_bytes(cert.get_notBefore()),
                "not_after": maybe_bytes(cert.get_notAfter()),
                "extensions": [],
            }

            for i in range(cert.get_extension_count()):
                ext = cert.get_extension(i)
                ext_info = {
                    "name": f"{maybe_bytes(ext.get_short_name())}",
                    "value": f"{maybe_bytes(ext)}",
                    "critical": ext.get_critical() != 0,
                }
                result["extensions"].append(ext_info)

            for entity, name in [
                ("subject", cert.get_subject()),
                ("issuer", cert.get_issuer()),
            ]:
                result[entity] = {}
                for key, value in name.get_components():
                    key = maybe_bytes(key)
                    value = maybe_bytes(value)
                    result[entity][f"{key}".lower()] = f"{value}"

            return result


def create_transport_details(
    transport: ITransport | IProcessTransport, is_server: bool
) -> TransportDetails:
    """
    Create transport details from Twisted transport.

    :param transport: The Twisted transport to extract information from.
    :param is_server: Flag indicating whether this transport side is a "server" (as in TCP server).
    :return: Transport details object filled with information from the Twisted transport.
    """
    peer = peer2str(transport)

    own_pid = os.getpid()
    if hasattr(threading, "get_native_id"):
        # New in Python 3.8
        # https://docs.python.org/3/library/threading.html?highlight=get_native_id#threading.get_native_id
        own_tid = threading.get_native_id()
    else:
        own_tid = threading.get_ident()
    own_fd = -1

    if _HAS_TLS and ISSLTransport.providedBy(transport):
        channel_id = {
            # this will only be filled when the TLS opening handshake is complete (!)
            "tls-unique": transport_channel_id(transport, is_server, "tls-unique"),
        }
        channel_type = TransportDetails.CHANNEL_TYPE_TLS
        peer_cert = extract_peer_certificate(transport)
        is_secure = True
    else:
        channel_id = {}
        channel_type = TransportDetails.CHANNEL_TYPE_TCP
        peer_cert = None
        is_secure = False

    # FIXME: really set a default (websocket)?
    channel_framing = TransportDetails.CHANNEL_FRAMING_WEBSOCKET

    td = TransportDetails(
        channel_type=channel_type,
        channel_framing=channel_framing,
        peer=peer,
        is_server=is_server,
        own_pid=own_pid,
        own_tid=own_tid,
        own_fd=own_fd,
        is_secure=is_secure,
        channel_id=channel_id,
        peer_cert=peer_cert,
    )

    return td



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

from autobahn import flatbuffers
from autobahn.flatbuffers.compat import import_numpy

np = import_numpy()


class Event(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Event()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsEvent(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Event
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Event
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Event
    def Subscription(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Event
    def Publication(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Positional values for application-defined event payload.
    # Event
    def Args(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Event
    def ArgsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Event
    def ArgsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Event
    def ArgsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Keyword values for application-defined event payload.
    # Event
    def Kwargs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Event
    def KwargsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Event
    def KwargsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Event
    def KwargsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Alternative, transparent payload. If given, ``args`` and ``kwargs`` must be left unset.
    # Event
    def Payload(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Event
    def PayloadAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Event
    def PayloadLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Event
    def PayloadIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # The specific scheme in use with Payload Passthru (PPT) mode for the application payload.
    # Event
    def PptScheme(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The specific serializer encoding the application payload with the Payload Passthru (PPT) scheme in use.
    # Event
    def PptSerializer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The cryptographic algorithm ("cipher") encrypting the application payload with the Payload Passthru (PPT) scheme in use.
    # Event
    def PptCipher(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The identifier or reference to the encryption key that was used to encrypt the payload with the Payload Passthru (PPT) scheme and cipher in use.
    # Event
    def PptKeyid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Event
    def Publisher(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Event
    def PublisherAuthid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Event
    def PublisherAuthrole(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Event
    def Topic(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Event
    def Retained(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # Event
    def TransactionHash(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Event
    def Acknowledge(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # Event
    def ForwardFor(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from autobahn.wamp.gen.wamp.proto.Principal import Principal

            obj = Principal()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Event
    def ForwardForLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Event
    def ForwardForIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        return o == 0


def EventStart(builder):
    builder.StartObject(18)


def Start(builder):
    EventStart(builder)


def EventAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    EventAddSession(builder, session)


def EventAddSubscription(builder, subscription):
    builder.PrependUint64Slot(1, subscription, 0)


def AddSubscription(builder, subscription):
    EventAddSubscription(builder, subscription)


def EventAddPublication(builder, publication):
    builder.PrependUint64Slot(2, publication, 0)


def AddPublication(builder, publication):
    EventAddPublication(builder, publication)


def EventAddArgs(builder, args):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(args), 0
    )


def AddArgs(builder, args):
    EventAddArgs(builder, args)


def EventStartArgsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartArgsVector(builder, numElems):
    return EventStartArgsVector(builder, numElems)


def EventAddKwargs(builder, kwargs):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(kwargs), 0
    )


def AddKwargs(builder, kwargs):
    EventAddKwargs(builder, kwargs)


def EventStartKwargsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartKwargsVector(builder, numElems):
    return EventStartKwargsVector(builder, numElems)


def EventAddPayload(builder, payload):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(payload), 0
    )


def AddPayload(builder, payload):
    EventAddPayload(builder, payload)


def EventStartPayloadVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayloadVector(builder, numElems):
    return EventStartPayloadVector(builder, numElems)


def EventAddPptScheme(builder, pptScheme):
    builder.PrependUint8Slot(6, pptScheme, 0)


def AddPptScheme(builder, pptScheme):
    EventAddPptScheme(builder, pptScheme)


def EventAddPptSerializer(builder, pptSerializer):
    builder.PrependUint8Slot(7, pptSerializer, 0)


def AddPptSerializer(builder, pptSerializer):
    EventAddPptSerializer(builder, pptSerializer)


def EventAddPptCipher(builder, pptCipher):
    builder.PrependUint8Slot(8, pptCipher, 0)


def AddPptCipher(builder, pptCipher):
    EventAddPptCipher(builder, pptCipher)


def EventAddPptKeyid(builder, pptKeyid):
    builder.PrependUOffsetTRelativeSlot(
        9, flatbuffers.number_types.UOffsetTFlags.py_type(pptKeyid), 0
    )


def AddPptKeyid(builder, pptKeyid):
    EventAddPptKeyid(builder, pptKeyid)


def EventAddPublisher(builder, publisher):
    builder.PrependUint64Slot(10, publisher, 0)


def AddPublisher(builder, publisher):
    EventAddPublisher(builder, publisher)


def EventAddPublisherAuthid(builder, publisherAuthid):
    builder.PrependUOffsetTRelativeSlot(
        11, flatbuffers.number_types.UOffsetTFlags.py_type(publisherAuthid), 0
    )


def AddPublisherAuthid(builder, publisherAuthid):
    EventAddPublisherAuthid(builder, publisherAuthid)


def EventAddPublisherAuthrole(builder, publisherAuthrole):
    builder.PrependUOffsetTRelativeSlot(
        12, flatbuffers.number_types.UOffsetTFlags.py_type(publisherAuthrole), 0
    )


def AddPublisherAuthrole(builder, publisherAuthrole):
    EventAddPublisherAuthrole(builder, publisherAuthrole)


def EventAddTopic(builder, topic):
    builder.PrependUOffsetTRelativeSlot(
        13, flatbuffers.number_types.UOffsetTFlags.py_type(topic), 0
    )


def AddTopic(builder, topic):
    EventAddTopic(builder, topic)


def EventAddRetained(builder, retained):
    builder.PrependBoolSlot(14, retained, 0)


def AddRetained(builder, retained):
    EventAddRetained(builder, retained)


def EventAddTransactionHash(builder, transactionHash):
    builder.PrependUOffsetTRelativeSlot(
        15, flatbuffers.number_types.UOffsetTFlags.py_type(transactionHash), 0
    )


def AddTransactionHash(builder, transactionHash):
    EventAddTransactionHash(builder, transactionHash)


def EventAddAcknowledge(builder, acknowledge):
    builder.PrependBoolSlot(16, acknowledge, 0)


def AddAcknowledge(builder, acknowledge):
    EventAddAcknowledge(builder, acknowledge)


def EventAddForwardFor(builder, forwardFor):
    builder.PrependUOffsetTRelativeSlot(
        17, flatbuffers.number_types.UOffsetTFlags.py_type(forwardFor), 0
    )


def AddForwardFor(builder, forwardFor):
    EventAddForwardFor(builder, forwardFor)


def EventStartForwardForVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartForwardForVector(builder, numElems):
    return EventStartForwardForVector(builder, numElems)


def EventEnd(builder):
    return builder.EndObject()


def End(builder):
    return EventEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from unittest.mock import Mock

import txaio

txaio.use_twisted()

from autobahn.twisted.websocket import (
    WebSocketClientProtocol,
    WebSocketServerFactory,
    WebSocketServerProtocol,
)
from autobahn.util import wildcards2patterns
from autobahn.wamp.types import TransportDetails
from autobahn.websocket.types import ConnectingRequest
from twisted.internet.error import ConnectionAborted, ConnectionDone, ConnectionLost
from twisted.python.failure import Failure
from twisted.trial import unittest

try:
    from twisted.internet.testing import StringTransport
except ImportError:
    from twisted.test.proto_helpers import StringTransport
from autobahn.testutil import FakeTransport


class ExceptionHandlingTests(unittest.TestCase):
    """
    Tests that we format various exception variations properly during
    connectionLost
    """

    def setUp(self):
        self.factory = WebSocketServerFactory()
        self.proto = WebSocketServerProtocol()
        self.proto.factory = self.factory
        self.proto.log = Mock()

    def tearDown(self):
        for call in [
            self.proto.autoPingPendingCall,
            self.proto.autoPingTimeoutCall,
            self.proto.openHandshakeTimeoutCall,
            self.proto.closeHandshakeTimeoutCall,
        ]:
            if call is not None:
                call.cancel()

    def test_connection_done(self):
        # pretend we connected
        self.proto._connectionMade()

        self.proto.connectionLost(Failure(ConnectionDone()))

        messages = " ".join([str(x[1]) for x in self.proto.log.mock_calls])
        self.assertTrue("closed cleanly" in messages)

    def test_connection_aborted(self):
        # pretend we connected
        self.proto._connectionMade()

        self.proto.connectionLost(Failure(ConnectionAborted()))

        messages = " ".join([str(x[1]) for x in self.proto.log.mock_calls])
        self.assertTrue(" aborted " in messages)

    def test_connection_lost(self):
        # pretend we connected
        self.proto._connectionMade()

        self.proto.connectionLost(Failure(ConnectionLost()))

        messages = " ".join([str(x[1]) for x in self.proto.log.mock_calls])
        self.assertTrue(" was lost " in messages)

    def test_connection_lost_arg(self):
        # pretend we connected
        self.proto._connectionMade()

        self.proto.connectionLost(Failure(ConnectionLost("greetings")))

        messages = " ".join([str(x[1]) + str(x[2]) for x in self.proto.log.mock_calls])
        self.assertTrue(" was lost " in messages)
        self.assertTrue("greetings" in messages)


class Hixie76RejectionTests(unittest.TestCase):
    """
    Hixie-76 should not be accepted by an Autobahn server.
    """

    def test_handshake_fails(self):
        """
        A handshake from a client only supporting Hixie-76 will fail.
        """
        t = FakeTransport()
        f = WebSocketServerFactory()
        p = WebSocketServerProtocol()
        p.factory = f
        p.transport = t

        # from http://tools.ietf.org/html/draft-hixie-thewebsocketprotocol-76
        http_request = b"GET /demo HTTP/1.1\r\nHost: example.com\r\nConnection: Upgrade\r\nSec-WebSocket-Key2: 12998 5 Y3 1  .P00\r\nSec-WebSocket-Protocol: sample\r\nUpgrade: WebSocket\r\nSec-WebSocket-Key1: 4 @1  46546xW%0l 1 5\r\nOrigin: http://example.com\r\n\r\n^n:ds[4U"

        p.openHandshakeTimeout = 0
        p._connectionMade()
        p.data = http_request
        p.processHandshake()
        self.assertIn(b"HTTP/1.1 400", t._written)
        self.assertIn(b"Hixie76 protocol not supported", t._written)


class WebSocketOriginMatching(unittest.TestCase):
    """
    Test that we match Origin: headers properly, when asked to
    """

    def setUp(self):
        self.factory = WebSocketServerFactory()
        self.factory.setProtocolOptions(
            allowedOrigins=["127.0.0.1:*", "*.example.com:*"]
        )
        self.proto = WebSocketServerProtocol()
        self.proto.transport = StringTransport()
        self.proto.factory = self.factory
        self.proto.failHandshake = Mock()
        self.proto._connectionMade()

    def tearDown(self):
        for call in [
            self.proto.autoPingPendingCall,
            self.proto.autoPingTimeoutCall,
            self.proto.openHandshakeTimeoutCall,
            self.proto.closeHandshakeTimeoutCall,
        ]:
            if call is not None:
                call.cancel()

    def test_match_full_origin(self):
        self.proto.data = b"\r\n".join(
            [
                b"GET /ws HTTP/1.1",
                b"Host: www.example.com",
                b"Sec-WebSocket-Version: 13",
                b"Origin: http://www.example.com.malicious.com",
                b"Sec-WebSocket-Extensions: permessage-deflate",
                b"Sec-WebSocket-Key: tXAxWFUqnhi86Ajj7dRY5g==",
                b"Connection: keep-alive, Upgrade",
                b"Upgrade: websocket",
                b"\r\n",  # last string doesn't get a \r\n from join()
            ]
        )
        self.proto.consumeData()

        self.assertTrue(self.proto.failHandshake.called, "Handshake should have failed")
        arg = self.proto.failHandshake.mock_calls[0][1][0]
        self.assertTrue("not allowed" in arg)

    def test_match_wrong_scheme_origin(self):
        # some monkey-business since we already did this in setUp, but
        # we want a different set of matching origins
        self.factory.setProtocolOptions(allowedOrigins=["http://*.example.com:*"])
        self.proto.allowedOriginsPatterns = self.factory.allowedOriginsPatterns
        self.proto.allowedOrigins = self.factory.allowedOrigins

        # the actual test
        self.factory.isSecure = False
        self.proto.data = b"\r\n".join(
            [
                b"GET /ws HTTP/1.1",
                b"Host: www.example.com",
                b"Sec-WebSocket-Version: 13",
                b"Origin: https://www.example.com",
                b"Sec-WebSocket-Extensions: permessage-deflate",
                b"Sec-WebSocket-Key: tXAxWFUqnhi86Ajj7dRY5g==",
                b"Connection: keep-alive, Upgrade",
                b"Upgrade: websocket",
                b"\r\n",  # last string doesn't get a \r\n from join()
            ]
        )
        self.proto.consumeData()

        self.assertTrue(self.proto.failHandshake.called, "Handshake should have failed")
        arg = self.proto.failHandshake.mock_calls[0][1][0]
        self.assertTrue("not allowed" in arg)

    def test_match_origin_secure_scheme(self):
        self.factory.isSecure = True
        self.factory.port = 443
        self.proto.data = b"\r\n".join(
            [
                b"GET /ws HTTP/1.1",
                b"Host: www.example.com",
                b"Sec-WebSocket-Version: 13",
                b"Origin: https://www.example.com",
                b"Sec-WebSocket-Extensions: permessage-deflate",
                b"Sec-WebSocket-Key: tXAxWFUqnhi86Ajj7dRY5g==",
                b"Connection: keep-alive, Upgrade",
                b"Upgrade: websocket",
                b"\r\n",  # last string doesn't get a \r\n from join()
            ]
        )
        self.proto.consumeData()

        self.assertFalse(
            self.proto.failHandshake.called, "Handshake should have succeeded"
        )

    def test_match_origin_documentation_example(self):
        """
        Test the examples from the docs
        """
        self.factory.setProtocolOptions(allowedOrigins=["*://*.example.com:*"])
        self.factory.isSecure = True
        self.factory.port = 443
        self.proto.data = b"\r\n".join(
            [
                b"GET /ws HTTP/1.1",
                b"Host: www.example.com",
                b"Sec-WebSocket-Version: 13",
                b"Origin: http://www.example.com",
                b"Sec-WebSocket-Extensions: permessage-deflate",
                b"Sec-WebSocket-Key: tXAxWFUqnhi86Ajj7dRY5g==",
                b"Connection: keep-alive, Upgrade",
                b"Upgrade: websocket",
                b"\r\n",  # last string doesn't get a \r\n from join()
            ]
        )
        self.proto.consumeData()

        self.assertFalse(
            self.proto.failHandshake.called, "Handshake should have succeeded"
        )

    def test_match_origin_examples(self):
        """
        All the example origins from RFC6454 (3.2.1)
        """
        # we're just testing the low-level function here...
        from autobahn.websocket.protocol import _is_same_origin, _url_to_origin

        policy = wildcards2patterns(["*example.com:*"])

        # should parametrize test ...
        for url in [
            "http://example.com/",
            "http://example.com:80/",
            "http://example.com/path/file",
            "http://example.com/;semi=true",
            # 'http://example.com./',
            "//example.com/",
            "http://@example.com",
        ]:
            self.assertTrue(
                _is_same_origin(_url_to_origin(url), "http", 80, policy), url
            )

    def test_match_origin_counter_examples(self):
        """
        All the example 'not-same' origins from RFC6454 (3.2.1)
        """
        # we're just testing the low-level function here...
        from autobahn.websocket.protocol import _is_same_origin, _url_to_origin

        policy = wildcards2patterns(["example.com"])

        for url in [
            "http://ietf.org/",
            "http://example.org/",
            "https://example.com/",
            "http://example.com:8080/",
            "http://www.example.com/",
        ]:
            self.assertFalse(_is_same_origin(_url_to_origin(url), "http", 80, policy))

    def test_match_origin_edge(self):
        # we're just testing the low-level function here...
        from autobahn.websocket.protocol import _is_same_origin, _url_to_origin

        policy = wildcards2patterns(["http://*example.com:80"])

        self.assertTrue(
            _is_same_origin(_url_to_origin("http://example.com:80"), "http", 80, policy)
        )
        self.assertFalse(
            _is_same_origin(_url_to_origin("http://example.com:81"), "http", 81, policy)
        )
        self.assertFalse(
            _is_same_origin(
                _url_to_origin("https://example.com:80"), "http", 80, policy
            )
        )

    def test_origin_from_url(self):
        from autobahn.websocket.protocol import _url_to_origin

        # basic function
        self.assertEqual(
            _url_to_origin("http://example.com"), ("http", "example.com", 80)
        )
        # should lower-case scheme
        self.assertEqual(
            _url_to_origin("hTTp://example.com"), ("http", "example.com", 80)
        )

    def test_origin_file(self):
        from autobahn.websocket.protocol import _url_to_origin

        self.assertEqual("null", _url_to_origin("file:///etc/passwd"))

    def test_origin_null(self):
        from autobahn.websocket.protocol import _is_same_origin, _url_to_origin

        self.assertEqual("null", _url_to_origin("null"))
        self.assertFalse(_is_same_origin(_url_to_origin("null"), "http", 80, []))
        self.assertFalse(_is_same_origin(_url_to_origin("null"), "https", 80, []))
        self.assertFalse(_is_same_origin(_url_to_origin("null"), "", 80, []))
        self.assertFalse(_is_same_origin(_url_to_origin("null"), None, 80, []))


class WebSocketXForwardedFor(unittest.TestCase):
    """
    Test that (only) a trusted X-Forwarded-For can replace the peer address.
    """

    def setUp(self):
        self.factory = WebSocketServerFactory()
        self.factory.setProtocolOptions(trustXForwardedFor=2)
        self.proto = WebSocketServerProtocol()
        self.proto.transport = StringTransport()
        self.proto.factory = self.factory
        self.proto.failHandshake = Mock()
        self.proto._connectionMade()

    def tearDown(self):
        for call in [
            self.proto.autoPingPendingCall,
            self.proto.autoPingTimeoutCall,
            self.proto.openHandshakeTimeoutCall,
            self.proto.closeHandshakeTimeoutCall,
        ]:
            if call is not None:
                call.cancel()

    def test_trusted_addresses(self):
        self.proto.data = b"\r\n".join(
            [
                b"GET /ws HTTP/1.1",
                b"Host: www.example.com",
                b"Origin: http://www.example.com",
                b"Sec-WebSocket-Version: 13",
                b"Sec-WebSocket-Extensions: permessage-deflate",
                b"Sec-WebSocket-Key: tXAxWFUqnhi86Ajj7dRY5g==",
                b"Connection: keep-alive, Upgrade",
                b"Upgrade: websocket",
                b"X-Forwarded-For: 1.2.3.4, 2.3.4.5, 111.222.33.44",
                b"\r\n",  # last string doesn't get a \r\n from join()
            ]
        )
        self.proto.consumeData()

        self.assertEqual(
            self.proto.peer,
            "2.3.4.5",
            "The second address in X-Forwarded-For should have been picked as the peer address",
        )


class OnConnectingTests(unittest.TestCase):
    """
    Tests related to onConnecting callback

    These tests are testing generic behavior, but are somewhat tied to
    'a framework' so we're just testing using Twisted-specifics here.
    """

    def test_on_connecting_client_fails(self):
        MAGIC_STR = "bad stuff"

        class TestProto(WebSocketClientProtocol):
            state = None
            wasClean = True
            log = Mock()

            def onConnecting(self, transport_details):
                raise RuntimeError(MAGIC_STR)

        proto = TestProto()
        proto.transport = FakeTransport()
        d = proto.startHandshake()
        self.successResultOf(d)  # error is ignored
        # ... but error should be logged
        self.assertTrue(len(proto.log.mock_calls) > 0)
        magic_found = False
        for i in range(len(proto.log.mock_calls)):
            if MAGIC_STR in str(proto.log.mock_calls[i]):
                magic_found = True
        self.assertTrue(magic_found, "MAGIC_STR not found when expected")

    def test_on_connecting_client_success(self):
        class TestProto(WebSocketClientProtocol):
            state = None
            wasClean = True
            perMessageCompressionOffers = []
            version = 18
            openHandshakeTimeout = 5
            log = Mock()

            def onConnecting(self, transport_details):
                return ConnectingRequest(
                    host="example.com",
                    port=443,
                    resource="/ws",
                )

        proto = TestProto()
        proto.transport = FakeTransport()
        proto.factory = Mock()
        proto._connectionMade()

        d = proto.startHandshake()
        req = self.successResultOf(d)
        self.assertEqual("example.com", req.host)
        self.assertEqual(443, req.port)
        self.assertEqual("/ws", req.resource)

    def test_str_transport(self):
        details = TransportDetails(
            channel_type=TransportDetails.CHANNEL_TYPE_FUNCTION,
            peer="example.com",
            is_secure=False,
            channel_id={},
        )
        # we can str() this and it doesn't fail
        str(details)

    def test_str_connecting(self):
        req = ConnectingRequest(host="example.com", port="1234", resource="/ws")
        # we can str() this and it doesn't fail
        str(req)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

"""
Centralized compiler flag selection for NVX (Native Vector Extensions) builds.

This module determines appropriate architecture-specific optimization flags
for building native C extensions using CFFI. It balances portability and
performance based on the build context.

Strategy
--------

**By default** - for *every* build context (PyPI wheels, local source installs,
and cross-compilation) - a safe, portable baseline architecture is used. This
keeps the resulting binaries runnable on a wide range of CPUs without SIGILL
(Illegal Instruction) crashes, and never hands a cross-compilation toolchain the
host-only ``-march=native`` flag (which it rejects).

Maximum-performance ``-march=native`` code generation is **opt-in** via
``AUTOBAHN_ARCH_TARGET=native``; use it only when the build host is also the run
host (e.g. Gentoo/Arch packages, dedicated single-machine deployments).

Architecture Baselines
----------------------

x86-64 (portable baseline):
    -march=x86-64-v2 (microarchitecture level 2, 2009+)
    Includes: SSE4.2, POPCNT, SSSE3, SSE4.1
    Compatible with: Intel Nehalem+, AMD Bulldozer+ (2009+)
    Coverage: ~99% of x86-64 CPUs from 2010 onwards

ARM64 (portable baseline):
    -march=armv8-a (baseline ARMv8-A architecture)
    Compatible with: Raspberry Pi 4/5, AWS Graviton, Apple M1/M2, etc.
    Coverage: All 64-bit ARM CPUs

Environment Variables
---------------------

AUTOBAHN_ARCH_TARGET : str, optional
    User/distro override for architecture target:
    - "native" : Force -march=native (maximum performance, build host only;
                 unsafe for distributed wheels and cross-compilation)
    - "safe"   : Force portable baseline (explicit; same as the default)
    - Not set  : Portable baseline (the safe default; works for cross-compilation)

AUTOBAHN_WHEEL_BUILD : str, optional
    Explicit marker for wheel builds ("true" or "1")
    When set, forces portable baseline regardless of auto-detection

CI : str, optional
    Standard CI environment variable. When "true" or "1", assumes wheel build.

CIBUILDWHEEL : str, optional
    Set by cibuildwheel. Indicates wheel build for distribution.

AUDITWHEEL_PLAT : str, optional
    Set by auditwheel/manylinux builds. Indicates wheel build.

Examples
--------

**GitHub Actions building wheels (default):**
    >>> # Default is the portable baseline
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=x86-64-v2']

**User installing from source (default):**
    >>> # Default is the portable baseline (safe for cross-compilation too)
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=x86-64-v2']

**Opting in to -march=native (build host == run host):**
    >>> import os
    >>> os.environ['AUTOBAHN_ARCH_TARGET'] = 'native'
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=native']

**Gentoo building package:**
    >>> # Gentoo wants -march=native for optimized binaries
    >>> import os
    >>> os.environ['AUTOBAHN_ARCH_TARGET'] = 'native'
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=native']

**Debian building package:**
    >>> # Debian wants portable binaries for all users
    >>> import os
    >>> os.environ['AUTOBAHN_ARCH_TARGET'] = 'safe'
    >>> get_compile_args()
    ['-std=c99', '-Wall', '-Wno-strict-prototypes', '-O3', '-march=x86-64-v2']

Notes
-----

This module is used by all NVX components:
- autobahn.nvx._xormasker (WebSocket frame XOR masking)
- autobahn.nvx._utf8validator (WebSocket UTF-8 validation)

Related Issues
--------------
- #1717: SIGILL crashes from -march=native in distributed wheels
- #1834: -march=native breaks cross-compilation; the default is now the portable
  baseline for all build contexts, with -march=native available opt-in.

See Also
--------
- https://en.wikipedia.org/wiki/X86-64#Microarchitecture_levels
- https://gcc.gnu.org/onlinedocs/gcc/x86-Options.html
"""

import os
import sys
import sysconfig
import platform


def is_building_wheel():
    """
    Detect if we're building a wheel for distribution vs. a local source install.

    .. note::

        As of #1834 the default architecture target is the portable baseline for
        *every* build context (wheels, local source installs, cross-compilation),
        so this helper no longer influences :func:`get_compile_args`. It is
        retained for backward compatibility and for external callers/tooling.

    Returns
    -------
    bool
        True if building a wheel (need portability), False if building locally.

    Notes
    -----
    Detection heuristics:
    1. CI environment variables (CI, CIBUILDWHEEL)
    2. manylinux/auditwheel markers (AUDITWHEEL_PLAT)
    3. Explicit marker (AUTOBAHN_WHEEL_BUILD)
    4. Default: False (assume local source build)
    """
    # CI environments (building wheels for distribution)
    if os.environ.get("CI") in ("true", "1"):
        return True

    # cibuildwheel
    if os.environ.get("CIBUILDWHEEL"):
        return True

    # manylinux/auditwheel builds
    if os.environ.get("AUDITWHEEL_PLAT"):
        return True

    # Explicit marker from our workflows
    if os.environ.get("AUTOBAHN_WHEEL_BUILD") in ("true", "1"):
        return True

    # Default: assume local source build
    return False


def get_compile_args():
    """
    Returns appropriate compiler arguments for building NVX native extensions.

    This function determines the optimal compiler flags based on:
    - Target platform (Windows, Linux, macOS, etc.)
    - Target architecture (x86-64, ARM64, etc.)
    - Build context (wheel distribution vs. local source install)
    - User/distro overrides via environment variables

    Returns
    -------
    list of str
        Compiler arguments suitable for current build context.
        For MSVC: ['/O2', '/W3']
        For GCC/Clang: ['-std=c99', '-Wall', '-O3', '-march=...']

    Examples
    --------
    >>> args = get_compile_args()
    >>> # Use with CFFI:
    >>> ffi.set_source("_nvx_module", c_source, extra_compile_args=args)
    """
    if sys.platform == "win32":
        # MSVC on Windows
        # Note: MSVC doesn't support -march, uses /arch instead
        # But default codegen is already optimized for current arch
        return ["/O2", "/W3"]

    # GCC/Clang on POSIX (Linux, macOS, *BSD)
    machine = _get_target_machine()

    # Base flags for all POSIX platforms
    base_args = [
        "-std=c99",
        "-Wall",
        "-Wno-strict-prototypes",
        "-O3",
    ]

    # User can explicitly override architecture target
    arch_override = os.environ.get("AUTOBAHN_ARCH_TARGET", "").lower()

    if arch_override == "native":
        # Explicit opt-in only: -march=native generates code for the exact CPU of
        # the *build* host (maximum performance). Use this when the build host is
        # also the run host (e.g. Gentoo, Arch Linux, performance-critical
        # deployments). It is NOT safe for distributed wheels or cross-compilation.
        return base_args + ["-march=native"]

    # Default for everyone (AUTOBAHN_ARCH_TARGET unset or "safe"): a portable
    # baseline architecture. Defaulting to "safe" rather than -march=native is
    # what makes cross-compilation work out of the box - a cross toolchain
    # rejects -march=native ("unknown value 'native' for '-march'", #1834) - and
    # also prevents SIGILL crashes from over-optimized distributed wheels (#1717).
    arch_flag = _get_safe_march_flag(machine)
    if arch_flag:
        return base_args + [arch_flag]

    # Unknown / cross target architecture: emit no -march flag and let the
    # toolchain defaults (or distro-supplied CFLAGS, e.g. Buildroot/Yocto)
    # govern code generation.
    return base_args


def _get_target_machine():
    """
    Return the *target* machine architecture for the build.

    Uses ``sysconfig.get_platform()`` rather than ``platform.machine()`` so that
    the correct architecture is detected when cross-compiling (e.g.
    Buildroot/Yocto building aarch64 on an x86-64 host): ``platform.machine()``
    reports the build host (``uname``), whereas ``sysconfig`` reflects the
    interpreter's configured target platform. Falls back to
    ``platform.machine()`` when no architecture token can be derived.

    (Target-detection approach contributed in PR #1835 by @jameshilliard.)

    Returns
    -------
    str
        Lower-cased target architecture token, e.g. "x86_64", "aarch64",
        "arm64".
    """
    plat = sysconfig.get_platform().lower()
    # Examples: 'linux-x86_64', 'linux-aarch64', 'macosx-11.0-arm64',
    #           'win-amd64', 'win32'.
    if plat == "win32":
        return "x86"
    machine = plat.rsplit("-", 1)[-1]
    return machine or platform.machine().lower()


def _get_safe_march_flag(machine):
    """
    Returns safe -march flag for portable binaries on the given machine architecture.

    Parameters
    ----------
    machine : str
        Machine architecture from platform.machine().lower()

    Returns
    -------
    str or None
        Safe -march flag for the architecture, or None for unknown architectures.
        Patched to always return None for Arch Linux x64 package builds.
        If Arch decides to increase CPU requirements [1], -march=x86-64-v3 or
        others could be used here.
        [1] https://gitlab.archlinux.org/archlinux/rfcs/-/blob/master/rfcs/0002-x86-64-v3-microarchitecture.md
    """
    if machine in ("x86_64", "amd64", "x64"):
        return None

    elif machine in ("aarch64", "arm64"):
        # ARM64: Use ARMv8-A baseline (all 64-bit ARM)
        # Compatible with: Raspberry Pi 4/5, AWS Graviton, Apple Silicon
        return "-march=armv8-a"

    else:
        # Unknown architecture: let compiler use safe defaults
        # (no -march flag = generic code for the architecture)
        return None



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# Key Derivation Functions, e.g. WAMP SCRAM uses a password-based key derivation function (KDF) to hash user passwords. WAMP-SCRAM supports both Argon2 and PBKDF2 as the KDF (see: WAMP protocol specification, "SCRAM Algorithms").
class KDF(object):
    # Not set / no TLS channel binding.
    NULL = 0
    # Argon2id variant of Argon2, version 1.3 - `argon2id13`.
    ARGON2ID13 = 1
    # PBKDF2 - `pbkdf2`.
    PBKDF2 = 2



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################


from autobahn.wamp.cryptosign import HAS_CRYPTOSIGN, CryptosignKey
from twisted.internet.defer import inlineCallbacks

__all__ = ["HAS_CRYPTOSIGN_SSHAGENT"]

if HAS_CRYPTOSIGN:
    try:
        # WAMP-cryptosign support for SSH agent is currently
        # only available on Twisted (on Python 2)
        from twisted.conch.ssh.agent import SSHAgentClient
        from twisted.internet.endpoints import UNIXClientEndpoint
        from twisted.internet.protocol import Factory
    except ImportError:
        # twisted.conch is not yet fully ported to Python 3
        HAS_CRYPTOSIGN_SSHAGENT = False
    else:
        HAS_CRYPTOSIGN_SSHAGENT = True
        __all__.append("SSHAgentCryptosignKey")


if HAS_CRYPTOSIGN_SSHAGENT:
    import os

    from nacl import signing

    from autobahn.wamp.cryptosign import _pack, _read_ssh_ed25519_pubkey, _unpack

    class SSHAgentCryptosignKey(CryptosignKey):
        """
        A WAMP-cryptosign signing key that is a proxy to a private Ed25510 key
        actually held in SSH agent.

        An instance of this class must be create via the class method new().
        The instance only holds the public key part, whereas the private key
        counterpart is held in SSH agent.
        """

        def __init__(self, key, comment=None, reactor=None):
            CryptosignKey.__init__(self, key, comment)
            if not reactor:
                from twisted.internet import reactor
            self._reactor = reactor

        @classmethod
        def new(cls, pubkey=None, reactor=None):
            """
            Create a proxy for a key held in SSH agent.

            :param pubkey: A string with a public Ed25519 key in SSH format.
            :type pubkey: unicode
            """
            if not HAS_CRYPTOSIGN_SSHAGENT:
                raise Exception(
                    "SSH agent integration is not supported on this platform"
                )

            pubkey, _ = _read_ssh_ed25519_pubkey(pubkey)

            if not reactor:
                from twisted.internet import reactor

            if "SSH_AUTH_SOCK" not in os.environ:
                raise Exception("no ssh-agent is running!")

            factory = Factory()
            factory.noisy = False
            factory.protocol = SSHAgentClient
            endpoint = UNIXClientEndpoint(reactor, os.environ["SSH_AUTH_SOCK"])
            d = endpoint.connect(factory)

            @inlineCallbacks
            def on_connect(agent):
                keys = yield agent.requestIdentities()

                # if the key is found in ssh-agent, the raw public key (32 bytes), and the
                # key comment as returned from ssh-agent
                key_data = None
                key_comment = None

                for blob, comment in keys:
                    raw = _unpack(blob)
                    algo = raw[0].decode("utf8")
                    if algo == "ssh-ed25519":
                        algo, _pubkey = raw
                        if _pubkey == pubkey:
                            key_data = _pubkey
                            key_comment = comment.decode("utf8")
                            break

                agent.transport.loseConnection()

                if key_data:
                    key = signing.VerifyKey(key_data)
                    return cls(key, key_comment, reactor)
                else:
                    raise Exception("Ed25519 key not held in ssh-agent")

            return d.addCallback(on_connect)

        def sign(self, challenge):
            if "SSH_AUTH_SOCK" not in os.environ:
                raise Exception("no ssh-agent is running!")

            factory = Factory()
            factory.noisy = False
            factory.protocol = SSHAgentClient
            endpoint = UNIXClientEndpoint(self._reactor, os.environ["SSH_AUTH_SOCK"])
            d = endpoint.connect(factory)

            @inlineCallbacks
            def on_connect(agent):
                # we are now connected to the locally running ssh-agent
                # that agent might be the openssh-agent, or eg on Ubuntu 14.04 by
                # default the gnome-keyring / ssh-askpass-gnome application
                blob = _pack([b"ssh-ed25519", self.public_key(binary=True)])

                # now ask the agent
                signature_blob = yield agent.signData(blob, challenge)
                algo, signature = _unpack(signature_blob)

                agent.transport.loseConnection()

                return signature

            return d.addCallback(on_connect)
